@@ -1,8 +1,8 @@
+from cgitb import handler
 import logging
 import asyncio
-
 import json
-from typing import Coroutine, Union
+from typing import Any, Coroutine, Mapping, Union
 
 from websockets import serve
 from websockets.exceptions import ConnectionClosedError, ConnectionClosedOK
@@ -11,36 +11,25 @@ from wpilib_ws import utils
 from wpilib_ws.hardware import DeviceType, CANDeviceType
 
 
-class InvalidDeviceError(Exception):
-    def __init__(self, device: str):
-        self.device = device
-        self.message = f"Device type '{device}' does not exists or cannot be handled."
-        super().__init__(self.message)
-
-
-class MessageEvent:
-
-    type: str
+class Message:
+    type: DeviceType | CANDeviceType
     device: str
-    data: dict
+    data: Mapping[str, Any]
 
-    def __init__(self, type: Union[DeviceType, CANDeviceType], device: str, data: dict):
+    def __init__(self, type, device, data):
         self.type = type
         self.device = device
-        self.payload = data
+        self.data = data
 
     @classmethod
-    def from_dict(cls, message: dict):
-        device_type = message["type"]
-        try:
-            if device_type.startswith("CAN"):
-                _type = CANDeviceType(device_type)
-            else:
-                _type = DeviceType(device_type)
+    def from_dict(cls, data: dict) -> "Message":
 
-            return cls(_type, message["device"], message["data"])
-        except ValueError:
-            raise InvalidDeviceError(device_type)
+        if data["type"].startswith("CAN"):
+            device_type = CANDeviceType(data["type"])
+        else:
+            device_type = DeviceType(data["type"])
+
+        return cls(device_type, data["device"], data["data"])
 
 
 class WPILibWsServer:
@@ -55,7 +44,8 @@ class WPILibWsServer:
         uri="/wpilibws",
         loop=None,
         logger=None,
-        debug=False,
+        verbose=False,
+        verbose_extreme=False,
     ):
         self._uri = uri
         self._connected = False
@@ -66,8 +56,9 @@ class WPILibWsServer:
         else:
             self._log = logger
 
-        self._debug = debug
-        if self._debug:
+        self._verbose_extreme = verbose_extreme
+        self._verbose = verbose or self._verbose_extreme
+        if self._verbose:
             self._log.setLevel(logging.DEBUG)
 
         self._handlers = {}
@@ -109,10 +100,15 @@ class WPILibWsServer:
         if self._connected:
             await ws.send("HTTP/1.1 409 Conflict\r\n")
             await ws.close(reason="Only one connection allowed at a time")
+            self._log.debug(
+                f"Client attempted to connect but was refused: {ws.remote_address[0]}:{ws.remote_address[1]}"
+            )
 
         try:
-            if path == "/wpilibws" and not self._connected:
-                self._log.info(f"Client connected: {ws.remote_address}")
+            if path == self._uri and not self._connected:
+                self._log.info(
+                    f"Client connected: {ws.remote_address[0]}:{ws.remote_address[1]}"
+                )
 
                 self._connected = True
                 self._ws = ws
@@ -124,9 +120,11 @@ class WPILibWsServer:
                     if not self.verify_data(data):
                         self._log.debug(f"Ignoring invalid data: {data}")
                     else:
-                        self._log.debug(f"Recieved message data: {data}")
-                        event = MessageEvent.from_dict(data)
-                        await self._handle_message(event)
+                        if self._verbose_extreme:
+                            self._log.debug(f"Recieved message data: {data}")
+                        message = Message.from_dict(data)
+                        await self._handle_message(message)
+
         except ConnectionClosedError as e:
             self._log.info(f"Socket Closed ({e.code}): {e.reason}")
         finally:
@@ -142,8 +140,12 @@ class WPILibWsServer:
     async def _send_ws_message(self, message):
         try:
             await self._ws.send(message)
+            if self._verbose_extreme:
+                self._log.debug(f"Sent message: {message}")
         except ConnectionClosedError:
-            self._log.debug("Could not send message because the socket was closed")
+            self._log.debug(
+                f"Could not send message because the socket was closed: {message}"
+            )
 
     def _start_background_tasks(self):
         self._loop.create_task(self._run_while_connected())
@@ -167,6 +169,15 @@ class WPILibWsServer:
         While robot code is connected, run the given coroutine in the
         background. This is best for sending payloads to the server for devices
         such as encoders, gyros, etc.
+
+        Example usage:
+        ```py
+        @server.while_connected()
+        async def while_connected():
+            await server.send_payload(
+                {"type": "RoboRIO", "device": "", "data": {">vin_voltage": 12.0}}
+            )
+        ```
         """
 
         def wrapper(coro):
@@ -175,16 +186,36 @@ class WPILibWsServer:
 
         return wrapper
 
-    def on_message(self, device_type: Union[str, DeviceType, CANDeviceType] = None):
+    def on_message(
+        self,
+        device_type: Union[str, DeviceType, CANDeviceType] = None,
+        device_name: str = "",
+    ):
         """
         Decorator for creating a message handler for all or a specific device
         type coming overthe websocket.
+
+        Example usage:
+        ```py
+        @server.on_message("PWM")
+        async def pwm_handler(message):
+            ...
+        ```
+        ```py
+        @server.on_message("SimDevice", "SPARK MAX")
+        async def spark_max_handler(message):
+            ...
+        ```
         """
         if isinstance(device_type, (DeviceType, CANDeviceType)):
             device_type = device_type.value
 
+        handler_name = device_type
+        if device_name:
+            handler_name += f":{device_name}"
+
         def wrapper(func):
-            event_name = device_type or "message"
+            event_name = handler_name or "message"
             self.add_handler(event_name, func)
             return func
 
@@ -209,13 +240,17 @@ class WPILibWsServer:
 
         self._background_tasks.append(wrapped)
 
-    async def _handle_message(self, event: MessageEvent):
+    async def _handle_message(self, event: Message):
         """
         Handle incoming messages, and call the appropriate handlers.
         """
 
+        names = (event.type.value, "message")
+
         for item in self._handlers.items():
-            if item[0] in (event.type.value, "message"):
+            if (
+                ":" in item[0] and event.device.startswith(item[0].split(":")[1])
+            ) or item[0] in names:
                 for func in item[1]:
                     await func(event)
 
@@ -225,7 +260,7 @@ class WPILibWsServer:
         """
         self._loop = asyncio.get_event_loop()
         async with serve(self._ws_handler, address, port, ping_interval=None) as s:
-            print(f"Listening on ws://{address}:{port}")
+            print(f"Listening on ws://{address}:{port}{self._uri}")
             await asyncio.Future()
 
     def run(self, address="0.0.0.0", port=3300):
